from django.conf import settings
from django.utils.encoding import force_text
from elasticsearch import NotFoundError
from elasticsearch_dsl.connections import connections
import elasticsearch_dsl as dsl

from .registry import model_documents

import importlib
import sys
import time


def import_class(fq_name):
    module_name, class_name = fq_name.rsplit('.', 1)
    mod = importlib.import_module(module_name)
    return getattr(mod, class_name)


def index(obj, index=None, using=None):
    """
    Shortcut to index a Django object based on it's model class.
    """
    from django.contrib.contenttypes.models import ContentType
    model_class = ContentType.objects.get_for_model(obj).model_class()
    for doc_class in model_documents.get(model_class, []):
        if not doc_class.queryset().filter(pk=obj.pk).exists():
            continue
        doc_using = using or doc_class._doc_type.using or 'default'
        doc_index = index or doc_class._doc_type.index or getattr(settings, 'SEEKER_INDEX', 'seeker')
        es = connections.get_connection(doc_using)
        body = doc_class.serialize(obj)
        doc_id = body.pop('_id', None)
        es.index(
            index=doc_index,
            doc_type=doc_class._doc_type.name,
            body=body,
            id=doc_id,
            refresh=True
        )


def delete(obj, index=None, using=None):
    """
    Shortcut to delete a Django object from the ES index based on it's model class.
    """
    from django.contrib.contenttypes.models import ContentType
    model_class = ContentType.objects.get_for_model(obj).model_class()
    for doc_class in model_documents.get(model_class, []):
        doc_using = using or doc_class._doc_type.using or 'default'
        doc_index = index or doc_class._doc_type.index or getattr(settings, 'SEEKER_INDEX', 'seeker')
        es = connections.get_connection(doc_using)
        try:
            es.delete(
                index=doc_index,
                doc_type=doc_class._doc_type.name,
                id=doc_class.get_id(obj),
                refresh=True
            )
        except NotFoundError:
            # If this object wasn't indexed for some reason (maybe not in the document's queryset), no big deal.
            pass


def search(models=None, using='default'):
    """
    Returns a search object across the specified models.
    """
    types = []
    indices = []
    if models is None:
        models = model_documents.keys()
    for model_class in models:
        for doc_class in model_documents.get(model_class, []):
            indices.append(doc_class._doc_type.index)
            types.append(doc_class)
    return dsl.Search(using=using).index(*indices).doc_type(*types)


def progress(iterator, count=None, label='', size=40, chars='# ', output=sys.stdout, frequency=1.0):
    """
    An iterator wrapper that writes/updates a progress bar to an output stream (stdout by default).
    Based on http://code.activestate.com/recipes/576986-progress-bar-for-console-programs-as-iterator/
    """
    assert len(chars) >= 2
    if label:
        label = force_text(label) + ' '

    try:
        count = len(iterator)
    except:
        pass

    start = time.time()

    def show(i):
        if count:
            x = int(size * i / count)
            bar = '[%s%s]' % (chars[0] * x, chars[1] * (size - x))
            pct = int((100.0 * i) / count)
            status = '%s/%s %s%%' % (i, count, pct)
        else:
            bar = ''
            status = str(i)
        e = time.time() - start
        mins, s = divmod(int(e), 60)
        h, m = divmod(mins, 60)
        elapsed = '%d:%02d:%02d' % (h, m, s) if h else '%02d:%02d' % (m, s)
        speed = '%.2f iters/sec' % (i / e) if e > 0 else ''
        output.write('%s%s %s - %s, %s\r' % (label, bar, status, elapsed, speed))
        output.flush()

    show(0)
    last_update = 0.0
    processed = 0
    for item in iterator:
        yield item
        processed += 1
        since = time.time() - last_update
        if since >= frequency:
            show(processed)
            last_update = time.time()
    show(processed)

    output.write('\n')
    output.flush()



from django.db import models

from .utils import delete, index

import logging
import warnings

logger = logging.getLogger(__name__)

class ModelIndexingMiddleware (object):
    """
    Deprecated: Middleware class that automatically indexes any new or deleted model objects.
    """

    def __init__(self, get_response=None):
        self.get_response = get_response
        warnings.warn("ModelIndexingMiddleware is deprecated for seeker. Please utilize seeker.SEEKER_INDEXER setting instead.", DeprecationWarning)

    def __call__(self, request):
        response = None
        if hasattr(self, 'process_request'):
            response = self.process_request(request)
        if not response:
            response = self.get_response(request)
        if hasattr(self, 'process_response'):
            response = self.process_response(request, response)
        return response

    def process_request(self, request):
        # This is really just here so Django keeps the middleware installed.
        pass



from django.dispatch import Signal

# TODO - let's rename this and make the sender the objects class (not instance)
search_complete = Signal(providing_args=('context'))
"""
Sent after a search is performed within a SeekerView object
The "sender" will be the instance of the SeekerView that was involved.
The "context" will be the context dictionary passed to the template. 
"""

advanced_search_performed = Signal(providing_args=['request', 'context', 'json_response'])
"""
Sent when an advanced search is executed. The following arguments will be provided:
- sender = The seeker view class that performed the search.
- request = The http request.
- context = The context dictionary that will be used to render this search.
- json_response = The json dictionary that will be returned as the response. 
                  This will include 'search_object' which defines the search that was performed.
"""


from .mapping import Indexable, ModelIndex

import logging
import threading


logger = logging.getLogger(__name__)


documents = []

model_documents = {}
model_doc_types = {}
app_documents = {}


def register(doc_class, app_label=None):
    assert issubclass(doc_class, Indexable)
    if doc_class in documents:
        logger.warning('Document class %s.%s was previously registered - skipping.', doc_class.__module__, doc_class.__name__)
        return
    documents.append(doc_class)
    if issubclass(doc_class, ModelIndex):
        model_class = doc_class.queryset().model
        # It's possible to register more than one document type for a model, so keep a list.
        model_documents.setdefault(model_class, []).append(doc_class)
        # For doing queries across multiple document types, we'll need a mapping from doc_type back to model_class.
        model_doc_types[doc_class._doc_type.name] = model_class
    if app_label:
        app_documents.setdefault(app_label, []).append(doc_class)



from django import template
from django.contrib.humanize.templatetags.humanize import intcomma
from django.core.paginator import Paginator
from django.template import loader
from django.utils.encoding import force_text
from django.utils.safestring import mark_safe
from six.moves.urllib.parse import parse_qsl, urlencode
import six

import datetime
import re


register = template.Library()

# Convenience so people don't need to install django.contrib.humanize
register.filter(intcomma)


@register.filter
def seeker_format(value):
    if value is None:
        return ''
    # TODO: settings for default list separator and date formats?
    if isinstance(value, datetime.datetime):
        return value.strftime('%m/%d/%Y %H:%M:%S')
    if isinstance(value, datetime.date):
        return value.strftime('%m/%d/%Y')
    if hasattr(value, '__iter__') and not isinstance(value, six.string_types):
        return ', '.join(force_text(v) for v in value)
    return force_text(value)


@register.filter
def seeker_filter_querystring(qs, keep):
    if isinstance(keep, six.string_types):
        keep = [keep]
    qs_parts = [part for part in parse_qsl(qs, keep_blank_values=True) if part[0] in keep]
    return urlencode(qs_parts)


@register.simple_tag
def seeker_facet(facet, results, selected=None, **params):
    params.update({
        'facet': facet,
        'selected': selected,
        'data': facet.data(results),
    })
    return loader.render_to_string(facet.template, params)

@register.simple_tag
def advanced_seeker_facet(facet, **params):
    """
    Renders an empty facet (no options/selections involved).
    The options/selected values should be set via AJAX.
    """
    params.update({
        'facet': facet,
    })
    return loader.render_to_string(facet.advanced_template, params)

@register.simple_tag
def seeker_column(column, result, **kwargs):
    return column.render(result, **kwargs)

@register.simple_tag
def seeker_column_header(column, results=None):
    return column.header(results)

@register.simple_tag
def seeker_score(result, max_score=None, template='seeker/score.html'):
    pct = result.meta.score / max_score if max_score else 0.0
    return loader.render_to_string(template, {
        'score': result.meta.score,
        'max_score': max_score,
        'percentile': pct * 100.0,
    })


@register.simple_tag
def seeker_pager(total, page_size=10, page=1, param='p', querystring='', spread=7, template='seeker/pager.html'):
    paginator = Paginator(range(total), page_size)
    if paginator.num_pages < 2:
        return ''
    page = paginator.page(page)
    if paginator.num_pages > spread:
        start = max(1, min(paginator.num_pages + 1 - spread, page.number - (spread // 2)))
        end = min(start + spread, paginator.num_pages + 1)
        page_range = range(start, end)
    else:
        page_range = paginator.page_range
    return loader.render_to_string(template, {
        'page_range': page_range,
        'paginator': paginator,
        'page': page,
        'param': param,
        'querystring': querystring,
    })

_phrase_re = re.compile(r'"([^"]*)"')


@register.simple_tag
def seeker_highlight(text, query, algorithm='english'):
    if not query:
        return mark_safe(seeker_format(text))
    try:
        import snowballstemmer
        stemmer = snowballstemmer.stemmer(algorithm)
        stemWord = stemmer.stemWord
        stemWords = stemmer.stemWords
    except:
        stemWord = lambda word: word
        stemWords = lambda words: words
    phrases = _phrase_re.findall(query)
    keywords = [w.lower() for w in re.split(r'\W+', _phrase_re.sub('', query)) if w]
    highlight = set(stemWords(keywords))
    text = seeker_format(text)
    for phrase in phrases:
        text = re.sub('(' + re.escape(phrase) + ')', r'<em>\1</em>', text, flags=re.I)
    parts = []
    for word in re.split(r'(\W+)', text):
        if stemWord(word.lower()) in highlight:
            parts.append('<em>%s</em>' % word)
        else:
            parts.append(word)
    return mark_safe(''.join(parts))



from django.contrib import admin

from .models import SavedSearch


class SavedSearchAdmin (admin.ModelAdmin):
    list_display = ('name', 'user', 'url', 'querystring', 'default', 'date_created')
    list_filter = ('url', 'user', 'default')

admin.site.register(SavedSearch, SavedSearchAdmin)



from django.db.models import signals

from .registry import model_documents
from .utils import delete, index

import logging
logger = logging.getLogger(__name__)

class ModelIndexer(object):
    """
    Class that automatically indexes any new or deleted mapped model objects.
    """

    def connect_signal_handlers(self):
        """
        Connects save and delete signal handler for mapped models. Also checks each ModelIndex for any additional signal handling that may be needed. 
        """

        for model_class, document_classes in model_documents.items():
            signals.post_save.connect(self.handle_save, sender=model_class, weak=False)
            signals.post_delete.connect(self.handle_delete, sender=model_class, weak=False)

            for document_class in document_classes:
                document_class.connect_additional_signal_handlers(self)

    def disconnect_signal_handlers(self):
        for model_class, document_classes in model_documents.items():
            signals.post_save.disconnect(self.handle_save, sender=model_class, weak=False)
            signals.post_delete.disconnect(self.handle_delete, sender=model_class, weak=False)

            for document_class in document_classes:
                document_class.disconnect_additional_signal_handlers(self)

    def handle_save(self, sender, instance, **kwargs):
        try:
            index(instance)
        except:
            logger.exception('Error indexing %s instance: %s', sender, instance)

    def handle_delete(self, sender, instance, **kwargs):
        try:
            delete(instance)
        except:
            logger.exception('Error deleting %s instance: %s', sender, instance)

    def handle_m2m_changed(self, sender, instance, action, **kwargs):
        if action in ('post_add', 'post_remove', 'post_clear'):
            try:
                index(instance)
            except:
                logger.exception('Error indexing many to many change %s instance: %s', sender, instance)





# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('seeker', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='savedsearch',
            name='user',
            field=models.ForeignKey(related_name='seeker_searches', to=settings.AUTH_USER_MODEL, on_delete=models.CASCADE),
        ),
    ]



import argparse

from django.conf import settings
from django.core.management.base import BaseCommand
from elasticsearch.helpers import bulk
from elasticsearch_dsl.connections import connections

from seeker.registry import app_documents, documents
from seeker.utils import progress


def reindex(doc_class, index, using, options):
    """
    Index all the things, using ElasticSearch's bulk API for speed.
    """
    def get_actions():
        for doc in doc_class.documents(cursor=options['cursor']):
            action = {
                '_index': index,
                '_type': doc_class._doc_type.name,
            }
            action.update(doc)
            yield action
    es = connections.get_connection(using)
    actions = get_actions() if options['quiet'] else progress(get_actions(), count=doc_class.count(), label=doc_class.__name__)
    bulk(es, actions)
    es.indices.refresh(index=index)


class Command (BaseCommand):
    args = '<app1 app2 ...>'
    help = 'Re-indexes the specified applications'

    def add_arguments(self, parser):
        parser.add_argument('--using',
            dest='using',
            default=None,
            help='The ES connection alias to use'
        )
        parser.add_argument('--index',
            dest='index',
            default=None,
            help='The ES index to store data in'
        )
        parser.add_argument('--quiet',
            action='store_true',
            dest='quiet',
            default=False,
            help='Do not produce any output while indexing')
        parser.add_argument('--drop',
            action='store_true',
            dest='drop',
            default=False,
            help='Drops the index before re-indexing'
        )
        parser.add_argument('--clear',
            action='store_true',
            dest='clear',
            default=False,
            help='Deletes all documents before re-indexing'
        )
        parser.add_argument('--no-data',
            action='store_false',
            dest='data',
            default=True,
            help='Only create the mappings, do not index any data'
        )
        parser.add_argument('--cursor',
            action='store_true',
            dest='cursor',
            default=False,
            help='Use a server-side cursor when fetching data for indexing')
        parser.add_argument('args', nargs=argparse.REMAINDER)

    def handle(self, *args, **options):
        doc_classes = []
        for label in args:
            doc_classes.extend(app_documents.get(label, []))
        if not args:
            doc_classes.extend(documents)
        if options['drop']:
            index = options['index'] or getattr(settings, 'SEEKER_INDEX', 'seeker')
            es = connections.get_connection(options['using'] or 'default')
            if es.indices.exists(index=index):
                es.indices.delete(index=index)
        for doc_class in doc_classes:
            using = options['using'] or doc_class._doc_type.using or 'default'
            index = options['index'] or doc_class._doc_type.index or getattr(settings, 'SEEKER_INDEX', 'seeker')
            if options['clear'] and not options['drop']:
                doc_class.clear(index=index, using=using)
            doc_class.init(index=index, using=using)
            if options['data']:
                reindex(doc_class, index, using, options)



from django.apps import apps
from django.core.management.base import BaseCommand
from elasticsearch.helpers import scan
from elasticsearch_dsl.connections import connections

from optparse import make_option
import json


class Command (BaseCommand):
    args = '<doc_type>'
    help = 'Dumps out data from the specified document types'
    option_list = BaseCommand.option_list + (
        make_option('--indent',
            type='int',
            dest='indent',
            default=None,
            help='Amount of indentation to use when serializing documents'
        ),
        make_option('--index',
            dest='index',
            default=None,
            help='Index to dump'
        ),
    )

    def handle(self, *args, **options):
        doc_types = ','.join(args) or None
        output = self.stdout
        output.write('[')
        es = connections.get_connection()
        for idx, doc in enumerate(scan(es, index=options['index'], doc_type=doc_types)):
            if idx > 0:
                output.write(',')
            output.write(json.dumps(doc, indent=options['indent']), ending='')
        output.write(']')
        output.flush()



from django.apps import apps
from django.core.management.base import BaseCommand, CommandError
from elasticsearch.helpers import bulk
from elasticsearch_dsl.connections import connections

from seeker.registry import model_doc_types

from optparse import make_option
import json


class Command (BaseCommand):
    args = '<app1 app2 ...>'
    option_list = BaseCommand.option_list + (
        make_option('--filename', '-f',
            dest='filename',
            default=None,
            help='The file to load index data from'
        ),
        make_option('--index',
            dest='index',
            default=None,
            help='Index to load data into'
        ),
    )

    def handle(self, *args, **options):
        if not options['filename']:
            raise CommandError('Please specify a file (-f) to read data from')

        refresh_indices = set()

        def get_actions():
            for data in json.load(open(options['filename'], 'rb')):
                if options['index']:
                    data['_index'] = options['index']
                refresh_indices.add(data['_index'])
                yield data

        es = connections.get_connection()
        bulk(es, get_actions())

        for index in refresh_indices:
            es.indices.refresh(index=index)



from django.conf import settings
from django.db import models
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible


@python_2_unicode_compatible
class SavedSearch (models.Model):
    user = models.ForeignKey(settings.AUTH_USER_MODEL, related_name='seeker_searches', on_delete=models.CASCADE)
    name = models.CharField(max_length=100, blank=True)
    url = models.CharField(max_length=200, db_index=True)
    querystring = models.TextField(blank=True)
    default = models.BooleanField(default=False)
    date_created = models.DateTimeField(default=timezone.now, editable=False)

    class Meta:
        ordering = ('name',)
        verbose_name_plural = 'saved searches'

    def __str__(self):
        return self.name

    def get_absolute_url(self):
        if self.querystring:
            return '%s?%s%s%s%d' % (self.url, self.querystring, ('&' if self.querystring else ''), 'saved_search=', self.pk)
        else:
            return self.url

    def get_details_dict(self):
        return { 'pk': self.pk, 'name': self.name, 'url': self.url, 'default': self.default }


@python_2_unicode_compatible
class AdvancedSavedSearch (SavedSearch):
    search_object = models.TextField()

    def __str__(self):
        return self.name

    def get_details_dict(self):
        details_dict = super(AdvancedSavedSearch, self).get_details_dict()
        details_dict.update({ 'search_object': self.search_object })
        return details_dict



__version__ = '4.1.4'

from .facets import Facet, GlobalTermsFacet, RangeFilter, TermsFacet, YearHistogram, DateTermsFacet, DateRangeFacet
from .mapping import (
    DEFAULT_ANALYZER, Indexable, ModelIndex, RawMultiString, RawString, build_mapping, deep_field_factory,
    document_field, document_from_model)
from .registry import app_documents, documents, model_documents, register
from .utils import delete, index, search
from .views import Column, AdvancedColumn, SeekerView, AdvancedSeekerView, AdvancedSavedSearchView


default_app_config = 'seeker.apps.SeekerConfig'
